import sys
import os
try:
    sys.path.insert(0, os.path.abspath('..'))  # works for local
    from spandas import Series
except:
    sys.path.insert(0, os.path.abspath('.'))  # works for local
    from spandas import Series
import pandas as pd
import numpy as np
import math

values = np.arange(5)
index = np.array(['a', 'b', 'c', 'd', 'e'])
a = Series(values, index)

def test_constructor():
    tmp = a
    assert np.all(tmp.values == np.arange(5)) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    # default as copy
    assert id(a.values) != id(values) and id(a.index) != id(index)
    tmp = Series(values, index, copy=False)
    assert id(tmp.values) == id(values) and id(tmp.index) == id(index)
    tmp = Series({'a': 1, 'b': 2})
    assert np.all(tmp.values == np.array([1, 2])) and \
           np.all(tmp.index == np.array(['a', 'b']))
    # initialize with pd.Series
    tmp_s = pd.Series(values, index)
    tmp = Series(tmp_s, tmp_s.index)
    assert np.all(tmp.values == np.arange(5)) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_indexing():
    tmp = a[3]
    assert tmp == 3
    tmp = a[np.array([1, 2, 3])]
    assert np.all(tmp.values == np.array([1, 2, 3])) and \
           np.all(tmp.index == np.array(['b', 'c', 'd']))
    tmp = a[1:3]
    assert np.all(tmp.values == np.array([1, 2])) and \
           np.all(tmp.index == np.array(['b', 'c']))
    tmp = a[a > 2]
    assert np.all(tmp.values == np.array([3, 4])) and \
           np.all(tmp.index == np.array(['d', 'e']))
    # get by index
    tmp = a[a.index == 'e']
    assert np.all(tmp.values == np.array([4])) and \
           np.all(tmp.index == np.array(['e']))
    tmp = a.get_by_index('e')
    assert np.all(tmp.values == np.array([4])) and \
           np.all(tmp.index == np.array(['e']))
    tmp = a.get_by_index(['b', 'c'])
    assert np.all(tmp.values == np.array([1, 2])) and \
           np.all(tmp.index == np.array(['b', 'c']))
    
    # test unchange
    assert np.all(a.values == np.arange(5)) and \
           np.all(a.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_copy():
    tmp = a.copy()
    assert np.all(tmp.values == a.values) and \
           np.all(tmp.index == a.index)
    # test unchange
    tmp.values[3] = 10
    assert np.all(a.values == np.arange(5)) and \
           np.all(a.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_assign():
    tmp = a.copy()
    tmp[1] = 10
    assert np.all(tmp.values == np.array([0, 10, 2, 3, 4])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a.copy()
    tmp[np.array([1, 2, 3])] = 10
    assert np.all(tmp.values == np.array([0, 10, 10, 10, 4])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a.copy()
    tmp[0:2] = 10
    assert np.all(tmp.values == np.array([10, 10, 2, 3, 4])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a.copy()
    # set by index
    tmp[a.index == 'e'] = 10
    assert np.all(tmp.values == np.array([0, 1, 2, 3, 10])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a.copy()
    tmp.set_by_index('e', 10)
    assert np.all(tmp.values == np.array([0, 1, 2, 3, 10])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a.copy()
    tmp.set_by_index(['e', 'd'], [10, 9])
    assert np.all(tmp.values == np.array([0, 1, 2, 9, 10])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    # test unchange
    assert np.all(a.values == np.arange(5)) and \
           np.all(a.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_iter():
    tmp = ""
    for i, v in a.items():
        tmp += "{}{}".format(i, v)
    assert tmp == "a0b1c2d3e4"
    tmp = 0
    for v in a:
        tmp += v
    assert tmp == 10
    
def test_len():
    assert len(a) == len(a.values) and len(a) == len(a.index)

def test_comp_op():
    tmp = a > 2
    assert np.all(tmp.values == np.array([False, False, False, True, True])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a < 2
    assert np.all(tmp.values == np.array([True, True, False, False, False])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a == Series([0, 1, 2, 2, 3])
    assert np.all(tmp.values == np.array([True, True, True, False, False])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_arithmetic_op():
    tmp = a + 1
    assert np.all(tmp.values == np.array([1, 2, 3, 4, 5])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    tmp = a - Series([1, 2, 3, 4, 5])
    assert np.all(tmp.values == np.array([-1, -1, -1, -1, -1])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
           
def test_dict():
    tmp = a.dict
    assert tmp == {'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4}
    
def test_append():
    tmp_append = Series([5])
    tmp = a.append(tmp_append)
    assert np.all(tmp.values == np.array([0, 1, 2, 3, 4, 5])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e', 0]))
    tmp = a.append([tmp_append, tmp_append])
    assert np.all(tmp.values == np.array([0, 1, 2, 3, 4, 5, 5])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e', 0, 0]))
    # test unchange
    assert np.all(a.values == np.arange(5)) and \
           np.all(a.index == np.array(['a', 'b', 'c', 'd', 'e']))

def test_apply():
    grade = Series([100, 90, 80, 60], index=['a', 'b', 'c', 'd'])
    tmp = grade.apply(lambda x: int(math.sqrt(x)*10))
    assert np.all(tmp.values == np.array([100, 94, 89, 77])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd']))
    tmp = grade.apply(np.mean)
    assert tmp == 82.5
    # test unchange
    assert np.all(grade.values == np.array([100, 90, 80, 60])) and \
           np.all(grade.index == np.array(['a', 'b', 'c', 'd']))

def test_map():
    tmp = a.map({i: i+3 for i in range(5)})
    assert np.all(tmp.values == np.array([3, 4, 5, 6, 7])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd', 'e']))
    # test unchange
    assert np.all(a.values == np.arange(5)) and \
           np.all(a.index == np.array(['a', 'b', 'c', 'd', 'e']))


"""
faster version of Series
"""
import numpy as np
import pandas as pd
"""
A Series is basically a data structure with 2 numpy array: 
    one to store value and the other to store index
"""
class Series:
    # initiate with pandas Series
    def __init__(self, values, index=None, copy=True):
        if isinstance(values, dict):
            if index is None:
                index = list(values.keys())
                values = list(values.values())
            else:
                values = [values[k] if k in values else np.nan for k in index]
        # try to convert values into a np array
        try:
            self.values = np.array(values, copy=copy)
        except:
            raise valuesError("values is not good")
        if index is not None:
            assert len(values) == len(index), \
                "length of values and index not match!"
            self.index = np.array(index, copy=copy)
        else:
            self.index = np.arange(len(values))

    def __str__(self):
        return str(pd.Series(self.values, self.index, copy=False))

    # len()
    def __len__(self):
        return len(self.values)

    """
    comparing operator
    just use the one for numpy
    """
    # >
    def __gt__(self, s):
        if isinstance(s, Series):
            tmp = self.values > s.values
        else:
            tmp = self.values > s
        return Series(tmp, self.index)

    # >=
    def __ge__(self, s):
        if isinstance(s, Series):
            tmp = self.values >= s.values
        else:
            tmp = self.values >= s
        return Series(tmp, self.index)

    # <
    def __lt__(self, s):
        if isinstance(s, Series):
            tmp = self.values < s.values
        else:
            tmp = self.values < s
        return Series(tmp, self.index)

    # <=
    def __le__(self, s):
        if isinstance(s, Series):
            tmp = self.values <= s.values
        else:
            tmp = self.values <= s
        return Series(tmp, self.index)

    # ==
    def __eq__(self, s):
        if isinstance(s, Series):
            tmp = self.values == s.values
        else:
            tmp = self.values == s
        return Series(tmp, self.index)

    # !=
    def __ne__(self, s):
        if isinstance(s, Series):
            tmp = self.values != s.values
        else:
            tmp = self.values != s
        return Series(tmp, self.index)
    
    """
    arithmetic
    """
    # +
    def __add__(self, s):
        if isinstance(s, Series):
            tmp = self.values + s.values
        else:
            tmp = self.values + s
        return Series(tmp, self.index)
    
    # -
    def __sub__(self, s):
        if isinstance(s, Series):
            tmp = self.values - s.values
        else:
            tmp = self.values - s
        return Series(tmp, self.index)
    
    # *
    def __mul__(self, s):
        if isinstance(s, Series):
            tmp = self.values * s.values
        else:
            tmp = self.values * s
        return Series(tmp, self.index)
    
    # /
    def __truediv__(self, s):
        if isinstance(s, Series):
            tmp = self.values / s.values
        else:
            tmp = self.values / s
        return Series(tmp, self.index)
    """
    indexing
    TODO: index slice
    """
    # all get does not copy
    def __getitem__(self, key):
        return self.get_by_row(key)
    
    def get_by_row(self, key):
        # boolean
        if isinstance(key, Series):
            assert len(self) == len(key), "the length of key does not match!"
            return Series(self.values[key.values], self.index[key.values], copy=False)
        try:
            return Series(self.values[key], self.index[key], copy=False)
        except:
            return self.values[key]
    
    def get_by_index(self, key):
        if not isinstance(key, list):
            key = [key]
        row2index = dict(zip(range(len(self)), self.index))
        key_row = [r for r, i in row2index.items() if i in key]
        return self[np.array(key_row, copy=False)]
    
    def __setitem__(self, key, value):
        self.set_by_row(key, value)

    def set_by_row(self, key, value):
        if isinstance(key, Series):
            assert len(self) == len(key), "the length of key does not match!"
            self.values[key.values] = value
        elif isinstance(key, list):
            key = np.array(key, copy=False)
        self.values[key] = value

    def set_by_index(self, key, value):
        try:
            k2v = dict(zip(key, value))
        except:
            k2v = {key: value}
        row2index = dict(zip(range(len(self)), self.index))
        key_row = [r for r, i in row2index.items() if i in k2v]
        value_row = [k2v[row2index[r]] for r in key_row]
        assert len(key_row) == len(value_row)
        self[np.array(key_row, copy=False)] = value_row
        
    def __iter__(self):
        for v in self.values:
            yield v
    
    def items(self):
        return zip(iter(self.index), iter(self.values))

    def append(self, s):
        # TODO: only support list of series now
        if isinstance(s, list):
            append_values = [l.values for l in s]
            append_index = [l.index for l in s]
            return Series(np.append(self.values, append_values), np.append(self.index, append_index))
        return Series(np.append(self.values, s.values), np.append(self.index, s.index))

    def apply(self, f, type='element'):
        if type == 'element':
            vf = np.vectorize(f)
            return Series(vf(self.values), self.index)
        elif type == 'column':
            return f(self.values)
        else:
            raise "type error!"

    def head(self, l = 5):
        return pd.Series(self.values, self.index, copy=False).head(l)

    def map(self, d, drop=True):
        if drop:
            return Series([d[x] if x in d else np.nan for x in self.values], self.index)
        else:
            return Series([d[x] if x in d else x for x in self.values], self.index)
    
    def copy(self):
        return Series(self.values, self.index)
    
    @property
    def dict(self):
        res = dict(zip(self.index, self.values))
        assert len(res) == len(self), "index have repetitive terms"
        return res
    
    def isnull(self):
        return Series(np.isnan(self.values), self.index)
    
    def to_pd(self, copy=True):
        return pd.Series(self.values, self.index, copy=copy)


import sys
import os
try:
    sys.path.insert(0, os.path.abspath('..'))  # works for local
    from spandas import DataFrame
except:
    sys.path.insert(0, os.path.abspath('.'))  # works for local
    from spandas import DataFrame
from spandas import Series
import pandas as pd
import numpy as np
import math

values = {'a': np.array([1, 2, 3]), 
          'b': np.array([4, 5, 6]), 
            1: np.array([6, 7, 8])}
index = np.array(['a', 'b', 'c'])
a = DataFrame(values, index)

def test_constructor():
    assert a.keys() == list(values.keys())
    tmp = DataFrame(values, index, copy=False)
    for k in values:
        assert id(values[k]) == id(tmp.dict[k])
    assert id(index) == id(tmp.index)
    # initialize with pd.DataFrame
    tmp_df = pd.DataFrame(values, index)
    tmp = DataFrame(tmp_df, tmp_df.index)
    for k in values:
        assert np.all(values[k] == tmp.dict[k])
    assert np.all(index == tmp.index)

def test_copy():
    tmp = a.copy()
    for k in values:
        assert id(values[k]) != id(tmp.dict[k])
    assert id(index) != id(tmp.index)

def test_indexing():
    # column
    tmp = a[['a']]
    assert isinstance(tmp, Series)
    assert np.all(tmp.values == np.array([1, 2, 3])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a[['a', 1]]
    assert isinstance(tmp, DataFrame)
    assert np.all(tmp.dict['a'] == np.array([1, 2, 3])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    # row
    tmp = a[2]
    assert np.all(tmp.dict['a'] == np.array([3])) and \
           np.all(tmp.dict['b'] == np.array([6])) and \
           np.all(tmp.dict[1] == np.array([8])) and \
           np.all(tmp.index == np.array(['c']))
    tmp = a[np.array([1, 2])]
    assert np.all(tmp.dict['a'] == np.array([2, 3])) and \
           np.all(tmp.dict['b'] == np.array([5, 6])) and \
           np.all(tmp.dict[1] == np.array([7, 8])) and \
           np.all(tmp.index == np.array(['b', 'c']))
    tmp = a[1:3]
    assert np.all(tmp.dict['a'] == np.array([2, 3])) and \
           np.all(tmp.dict['b'] == np.array([5, 6])) and \
           np.all(tmp.dict[1] == np.array([7, 8])) and \
           np.all(tmp.index == np.array(['b', 'c']))
    # get by index
    tmp = a.get_by_index(['a', 'c'])
    assert np.all(tmp.dict['a'] == np.array([1, 3])) and \
           np.all(tmp.dict['b'] == np.array([4, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 8])) and \
           np.all(tmp.index == np.array(['a', 'c']))
    # test unchange
    tmp = a
    assert np.all(tmp.dict['a'] == np.array([1, 2, 3])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))

def test_assign():
    # column assignment
    tmp = a.copy()
    tmp[['a', 'c']] = 1
    assert np.all(tmp.dict['a'] == np.array([1, 1, 1])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.dict['c'] == np.array([1, 1, 1])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a.copy()
    tmp[['a', 'c']] = [3, 2, 1]
    assert np.all(tmp.dict['a'] == np.array([3, 2, 1])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.dict['c'] == np.array([3, 2, 1])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    # row assignment
    tmp = a.copy()
    tmp[0] = 30
    assert np.all(tmp.dict['a'] == np.array([30, 2, 3])) and \
           np.all(tmp.dict['b'] == np.array([30, 5, 6])) and \
           np.all(tmp.dict[1] == np.array([30, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a.copy()
    tmp[:2] = 30
    assert np.all(tmp.dict['a'] == np.array([30, 30, 3])) and \
           np.all(tmp.dict['b'] == np.array([30, 30, 6])) and \
           np.all(tmp.dict[1] == np.array([30, 30, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a.copy()
    tmp.set_by_index(['a', 'c'], [20, 30])
    assert np.all(tmp.dict['a'] == np.array([20, 2, 30])) and \
           np.all(tmp.dict['b'] == np.array([20, 5, 30])) and \
           np.all(tmp.dict[1] == np.array([20, 7, 30])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    # nested assignment
    tmp = a.copy()
    tmp[['b']][:2] = 30
    assert np.all(tmp.dict['a'] == np.array([1, 2, 3])) and \
           np.all(tmp.dict['b'] == np.array([30, 30, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a.copy()
    tmp[['a', 'b']][:2] = 30
    assert np.all(tmp.dict['a'] == np.array([30, 30, 3])) and \
           np.all(tmp.dict['b'] == np.array([30, 30, 6])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))
    tmp = a.copy()
    tmp[['a', 'b']].set_by_index('c', 30)
    assert np.all(tmp.dict['a'] == np.array([1, 2, 30])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 30])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c']))

def test_transpose():
    tmp = a.T
    assert np.all(tmp.dict['a'] == np.array([1, 4, 6])) and \
           np.all(tmp.dict['b'] == np.array([2, 5, 7])) and \
           np.all(tmp.dict['c'] == np.array([3, 6, 8])) and \
           np.all(tmp.index == np.array(['a', 'b', 1]))

def test_len():
    assert len(a) == 3

def test_columns():
    assert a.columns == ['a', 'b', 1]

def test_append():
    tmp_append = DataFrame({'a': [0], 'b': [1], 1: [2]})
    tmp = a.append(tmp_append)
    assert np.all(tmp.dict['a'] == np.array([1, 2, 3, 0])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 6, 1])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8, 2])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 0]))
    tmp = a.append([tmp_append, tmp_append])
    assert np.all(tmp.dict['a'] == np.array([1, 2, 3, 0, 0])) and \
           np.all(tmp.dict['b'] == np.array([4, 5, 6, 1, 1])) and \
           np.all(tmp.dict[1] == np.array([6, 7, 8, 2, 2])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 0, 0]))

def test_iter():
    tmp = ""
    for k in a:
        tmp += str(k)
    assert tmp == "ab1"
    tmp = ""
    tmp_sum = 0
    for k, v in a.items():
        tmp += str(k)
        tmp_sum += v.sum()
    assert tmp == "ab1" and tmp_sum == 42
    tmp = ""
    tmp_sum = 0
    for k, v in a.iterrows():
        tmp += str(k)
        tmp_sum += v['b']
    assert tmp == "abc" and tmp_sum == 15

def test_apply():
    grade = DataFrame({'A': [100, 90, 80, 60],
                           'B': [90, 85, 85, 100],
                           'C': [70, 100, 100, 80]},
                          index=['a', 'b', 'c', 'd'])
    # element-wise
    tmp = grade.apply(lambda x: int(math.sqrt(x)*10))
    assert np.all(tmp.dict['A'] == np.array([100, 94, 89, 77])) and \
           np.all(tmp.dict['B'] == np.array([94, 92, 92, 100])) and \
           np.all(tmp.dict['C'] == np.array([83, 100, 100, 89])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd']))
    # row-wise
    tmp = grade.apply(lambda x: x['A']+x['B']+x['C'], type='row')
    assert np.all(tmp.values == np.array([260, 275, 265, 240])) and \
           np.all(tmp.index == np.array(['a', 'b', 'c', 'd']))
    # column wise
    tmp = grade.apply(lambda x: int(x.mean()), type='column')
    assert np.all(tmp.values == np.array([82, 90, 87])) and \
           np.all(tmp.index == np.array(['A', 'B', 'C']))
    # test unchanged
    assert np.all(grade.dict['A'] == np.array([100, 90, 80, 60])) and \
           np.all(grade.dict['B'] == np.array([90, 85, 85, 100])) and \
           np.all(grade.dict['C'] == np.array([70, 100, 100, 80])) and \
           np.all(grade.index == np.array(['a', 'b', 'c', 'd']))


from .series import Series
from collections.abc import Iterable
import numpy as np
import pandas as pd
import random

class DataFrame:
    def __init__(self, values, index=None, columns=None, copy=True):
        self.dict = {}
        l = 0
        try:
            if columns is not None:
                for key in columns:
                    try:
                        value = values[key]
                    except:
                        continue
                    if l == 0:
                        l = len(value)
                    else:
                        assert l == len(value), "the length of each column must match!"
                    self.dict[key] = np.array(value, copy=copy)
                for key in columns:
                    if key not in self.dict:
                        self.dict[key] = np.empty(l)
                        self.dict[key][:] = np.nan
            else:
                for key, value in values.items():
                    if l == 0:
                        l = len(value)
                    else:
                        assert l == len(value), "the length of each column must match!"
                    self.dict[key] = np.array(value, copy=copy)
        except:
            for key, value in values.items():
                self.dict[key] = np.array([value], copy=copy)
        if index is not None:
            assert l == len(index), "the length of each column must match!"
            self.index = np.array(index, copy=copy)
        else:
            self.index = np.array(range(l), copy=copy)

    def __str__(self):
        return str(pd.DataFrame(self.dict, self.index, copy=False))

    def head(self, l = 5):
        return pd.DataFrame(self.dict, self.index, copy=False).head(l)

    def __len__(self):
        return len(self.index)

    def __getitem__(self, key):
        # index with column name
        if isinstance(key, list):
            if len(key) == 1:
                return Series(self.dict[key[0]], self.index, copy=False)
            else:
                return DataFrame({k: self.dict[k] for k in key}, self.index, copy=False)
        # index with boolean Series
        elif isinstance(key, Series):
            assert len(key) == len(self)
            return DataFrame({k: v[key.values] for k, v in self.dict.items()}, self.index[key.values], copy=False)
        else:
            if not isinstance(key, Iterable) and not isinstance(key, slice):
                key = [key]
            return DataFrame({k: v[key] for k, v in self.dict.items()}, self.index[key], copy=False)
    
    def get_by_index(self, key):
        if not isinstance(key, list):
            key = [key]
        row2index = dict(zip(range(len(self)), self.index))
        key_row = [r for r, i in row2index.items() if i in key]
        return self[np.array(key_row, copy=False)]
    
    # when assigning with column and row, 
    # please always use the row first and column second
    # because numpy advance indexing will return a copy instead for a view
    def __setitem__(self, key, value):
        # index with column name
        if isinstance(key, list):
            if isinstance(value, Series):
                value = value.values
            for k in key:
                if k in self.dict:
                    self.dict[k][:] = value
                else:  # add new column
                    self.dict[k] = np.empty(len(self))
                    self.dict[k][:] = value
        # index with boolean Series
        elif isinstance(key, Series):
            for k in self:
                self.dict[k][key.values] = value
        else:
            if not isinstance(key, Iterable) and not isinstance(key, slice):
                key = np.array([key])
            for k in self:
                self.dict[k][key] = value
    
    def set_by_index(self, key, value):
        try:
            k2v = dict(zip(key, value))
        except:
            k2v = {key: value}
        row2index = dict(zip(range(len(self)), self.index))
        key_row = [r for r, i in row2index.items() if i in k2v]
        value_row = [k2v[row2index[r]] for r in key_row]
        assert len(key_row) == len(value_row)
        self[np.array(key_row, copy=False)] = value_row

    def __delitem__(self, key):
        try:
            assert not isinstance(key, str)
            for k in key:
                try:
                    del self.dict[k]
                except:
                    pass
        except:
            del self.dict[key]

    def __iter__(self):
        for k in self.dict:
            yield k

    def items(self):
        for k, v in self.dict.items():
            yield k, v

    def iterrows(self):
        for i in range(len(self.index)):
            yield self.index[i], {k: v[i] for k, v in self.dict.items()}

    def append(self, s):
        # only support append list of DataFrame or dict now
        if isinstance(s, list):
            append_dict = {k: [] for k in self.keys()}
            append_index = []
            for d in s:
                assert d.keys() == self.keys(), "keys not conform"
                if not isinstance(d, DataFrame):
                    d = DataFrame(d)
                for k in append_dict.keys():
                    append_dict[k].append(d.dict[k])
                append_index.append(d.index)
            res_dict = {k: np.append(self.dict[k], append_dict[k]) for k in self.keys()}
            res_index = np.append(self.index, append_index)
            return DataFrame(res_dict, res_index)
        else:
            if not isinstance(s, DataFrame):
                s = DataFrame(s)
            res_dict = {k: np.append(self.dict[k], s.dict[k]) for k in self.keys()}
            res_index = np.append(self.index, s.index)
            return DataFrame(res_dict, res_index)

    def keys(self):
        return list(self.dict.keys())
    
    def copy(self):
        return DataFrame(self.dict, self.index)
    
    @property
    def columns(self):
        return list(self.dict.keys())
    
    @property
    def T(self):
        res_index = self.columns
        res_columns = self.index
        res_values = np.array([self.dict[k] for k in res_index], copy=False).T
        res_dict = dict(zip(res_columns, res_values))
        return DataFrame(res_dict, index=res_index, columns=res_columns)

    def apply(self, func, type='element'):
        if type == 'element':
            f = np.vectorize(func)
            res_dict = {k: f(v) for k, v in self.dict.items()}
            return DataFrame(res_dict, self.index)
        elif type == 'row':
            res_values = []
            for _, v in self.iterrows():
                res_values.append(func(v))
            return Series(res_values, self.index)
        elif type == 'column':
            return Series([func(v) for _, v in self.items()], self.keys())
    
    def sample(self, n):
        perm = np.array(random.sample(range(len(self)), n), copy=False)
        return DataFrame({k: v[perm] for k, v in self.items()}, self.index[perm])
    
    def to_pd(self, copy=True):
        return pd.DataFrame(self.dict, self.index, copy=copy)